unit PreProcessorU;
// DONE 5 : Improve pre-processor speed
{.$DEFINE BreakOnError}
{$DEFINE CaseSensitive}
interface
uses
  sysutils,
  classes,
  contnrs,
  ScannerU;

Type
  TCaseSensitiveStringList = class( TStringList )
  public
    function Find( const S : string; var Index : Integer ) : Boolean; override;
    function IndexOf( const S : string ) : Integer; override;
  end;


  EPreProcessorError = class(Exception);

  TTokenType = (
                tt_Eof,
                tt_Ident,
                tt_Other,
                tt_Define
               );

  TPreProcessorParser = class;
  TPreProcessorDefineStates = class;
  TPreProcessorDefineState = class;
  TPreProcessorDefine = class;
  TPreProcessor = class;

  TOnNewLine = procedure ( Sender : TPreProcessorParser;
                           Row, Column, Position : integer ) of object;

  TOnMacroExpansion = function ( Sender : TPreProcessorParser;
                                 const MacroName : string ) : string of object;

  TPreProcessorParser = Class
  Private
    fData : string;
    fText : pchar;
    fToken : string;
    fTokenType : TTokenType;
    fLastEnterPos : integer;
    fLength : integer;
    fRow,fColumn,fPosition : integer;
    fOnNewLine: TOnNewLine;
    fOnMacroExpansion : TOnMacroExpansion;
  public
    procedure SetText(const data: string);
    procedure Next;
    
    property Token: string read fToken;
    property TokenType : TTokenType read fTokenType;

    property Row : integer read fRow;
    property Column : integer read fColumn;
    property Position : integer read fPosition;

    property OnNewLine : TOnNewLine read fOnNewLine write fOnNewLine;
    property OnMacroExpansion : TOnMacroExpansion read fOnMacroExpansion write fOnMacroExpansion;
  end; {TPreProcessorParser}

  TPreProcessorDefine = class
  protected
    fName : string;
    fValue : string;
    fGlobalDefine : boolean;
  public
    Constructor Create( const aName, aValue : string );
    property Name : string read fName write fName;
    property Value : string read fValue write fValue;    
  end; {TPreProcessorDefine}

  TPreProcessorDefineState = class
  private
    fInElse: Boolean;
    fDoWrite: Boolean;
  public
    property InElse: Boolean read fInElse write fInElse;
    property DoWrite: Boolean read fDoWrite write fDoWrite;
  end; {TPreProcessorDefineState}

  TPreProcessorDefineStates = class
  private
    fItems : Tlist;
    function GetCount : integer;
    function GetItem( Index : integer ) : TPreProcessorDefineState;
    function GetWrite : Boolean;
  public
    constructor Create;
    destructor Destroy; override;

    function Add : TPreProcessorDefineState;
    procedure Delete( index : integer );

    property Count: Longint read GetCount;
    property Item[ index : integer ]: TPreProcessorDefineState read GetItem; default;

    property DoWrite: Boolean read GetWrite;
  end; {TPreProcessorDefineStates}


  TLineInfo = class
  protected
    fEndPos : integer;
    fStartPos : integer;
    fFileName : string;
    fLineOffsets : TList;

    function GetLineOffset( index : Integer ): integer;
    function GetLineOffsetCount: Longint;
  public
    constructor Create;
    destructor Destroy; override;

    property FileName : string read fFileName;
    property StartPos : integer read fStartPos;
    property EndPos : integer read fEndPos;
    property LineOffsetCount: integer read GetLineOffsetCount;
    property LineOffset[ index : integer] : integer read GetLineOffset;
  end; {TLineInfo}

  TLineInfoResults = record
    Row : integer;
    Column : integer;
    Position : integer;
    Name : string;
  end; {TLineInfoResults}

  TLineInfoList = class
  protected
    fItems : TList;
    fCurrent : integer;

    function GetCount : integer;
    function GetItem( index : integer): TLineInfo;

    function Add: TLineInfo;
  public
    constructor Create;
    destructor Destroy; override;

    procedure Clear;

    function GetLineInfo( Position : integer;
                          var Res: TLineInfoResults ): Boolean;
                          
    property Current : integer read fCurrent write FCurrent;
    property Count : integer read GetCount;
    property Items[ index : integer] : TLineInfo read GetItem; default;
  end; {TLineInfoList}

  TOnCommand = function ( Sender: TPreProcessor;
                          const command, param : string): boolean of object;

  TOnNeedFile = function ( Sender: TPreProcessor;
                           const CallingFilename: string;
                           var FileName, Output: string): Boolean of object;
  TPreProcessor = class
  protected
    fMainFileData : string;
    fMainFileName : string;
    fMaxLevel : integer;
    fAddedPosition : integer;
    fDefinesOwner : TObjectList;
    fCurrentDefinesOwner : TObjectList;
    FCurrentLineInfo: TLineInfoList;

    fDefines : TStringList;
    // list of all defines
    fCurrentDefines : TStringList;
    // active states
    fDefineState : TPreProcessorDefineStates;

    fOnNeedFile : TOnNeedFile;
    fOnCommand : TOnCommand;
    function OnMacroExpansion( Sender : TPreProcessorParser;
                               const MacroName : string ) : string;

    procedure ParserNewLine( Sender: TPreProcessorParser;
                             Row, Column, Position: integer);
                                                               
    procedure PreProcessFile( Level: Integer;
                              const OrgFileName: string;
                              FileName: string;
                              Dest: TStream);
  public
    constructor Create;
    destructor Destroy; override;

    procedure PreProcess( const Filename: string;
                          var aOutput: string);

    // does not clear global defines
    procedure Clear;

    Procedure AddPrivateDefine( const Name, value : string );
    Procedure RemovePrivateDefine( const Name : string );
    Function PrivateDefineExists( const Name : string ) : boolean;

        
    Procedure AddDefine( const Name, value : string );
    Procedure RemoveDefine( const Name : string );
    Function DefineExists( const Name : string ) : boolean;
    Procedure ClearDefines;

    // active states
    Property DefineState : TPreProcessorDefineStates read fDefineState;

    // maxium number of includes allowed
    property MaxLevel: integer read fMaxLevel write fMaxLevel; // default 20;
    property CurrentLineInfo : TLineInfoList read FCurrentLineInfo;
    property OnNeedFile : TOnNeedFile read FOnNeedFile write FOnNeedFile;
    Property OnCommand : TOnCommand read fOnCommand write fOnCommand;
    // list of global defines
    property Defines : TStringList read FDefines write FDefines;
    // the data for the main file
    property MainFileData : string read fMainFileData write fMainFileData;
    // the name of the main file
    property MainFileName : string read fMainFileName write fMainFileName;
  end; {TPreProcessor}



Procedure ParseParams( Params : string;
                       var Param1 : string;
                       var Param2 : string );

function UnQuote( const s : string) : string;
  
implementation


// -----------------------------------------------------------------------------
//   TCaseSensitiveStringList
// -----------------------------------------------------------------------------

function TCaseSensitiveStringList.IndexOf(const S: string): Integer;
begin
if not Sorted then
  begin
  for Result := 0 to GetCount - 1 do
    if AnsiCompareStr(Get(Result), S) = 0 then Exit;
  Result := -1;
  end
else if not Find(S, Result) then Result := -1;
end;

function TCaseSensitiveStringList.Find(const S: string; var Index: Integer): Boolean;
var
  L, H, I, C: Integer;
begin
  Result := False;
  L := 0;
  H := Count - 1;
  while L <= H do
  begin
    I := (L + H) shr 1;
    C := AnsiCompareStr(self[I], S);
    if C < 0 then L := I + 1 else
    begin
      H := I - 1;
      if C = 0 then
      begin
        Result := True;
        {if list.Duplicates <> dupAccept then }L := I;
      end;
    end;
  end;
  Index := L;
end; {Find}

// -----------------------------------------------------------------------------
//   TPreProcessorParser
// -----------------------------------------------------------------------------

procedure TPreProcessorParser.SetText(const data: string);
begin
fData := data;
fText := pchar(fData);
fLength := 0;
fPosition := 0;
fColumn := 1;
FLastEnterPos := 0;
fRow := 1;
if @fOnNewLine <> nil then
  fOnNewLine(Self, 1, 1, 0);
Next;
end; {SetText}

procedure TPreProcessorParser.Next;
var
  tmpPosition : integer;
begin
inc( fPosition, fLength );
fLength := 0;
case fText[ fPosition ] of
  #0 : // end of file
    begin
    fLength := 0;
    fTokenType := tt_Eof;
    end;
  '/': // comment
    begin
    tmpPosition := fPosition + 1;
    if fText[ tmpPosition ] = '*' then
      begin
      while (FText[ tmpPosition ] <> #0) do
        begin
        if ( fText[ tmpPosition ] = '*' ) and
           ( fText[ tmpPosition + 1] = '/' )  then
          Break
        else if fText[ tmpPosition ] = #13 then
          begin
          inc( FRow );
          if FText[ tmpPosition + 1 ] = #10 then
            inc( tmpPosition );
          FLastEnterPos := fPosition - 1;
          if @fOnNewLine <> nil then
            OnNewLine( Self, FRow, fPosition - FLastEnterPos + 1,
                       tmpPosition + 1);
          end
        else if fText[ tmpPosition ] = #10 then
          begin
          inc( fRow );
          FLastEnterPos := tmpPosition - 1;
          if @fOnNewLine <> nil then
            OnNewLine( Self, FRow, fPosition - FLastEnterPos + 1,
                       tmpPosition + 1);
          end;
        Inc(tmpPosition);
        end;
      fTokenType := tt_Other;
      if (FText[tmpPosition] <> #0) then
        Inc(tmpPosition, 2);
      fLength := tmpPosition - fPosition;
      end
    else if fText[ tmpPosition ] = '/' then
      begin
      while not ( fText[ tmpPosition ] in [ #0, #13, #10 ] ) do
        Inc( tmpPosition );
      fTokenType := tt_Other;
      fLength := tmpPosition - fPosition;
      end
    else
      begin
      fTokenType := tt_Other;
      fLength := 1;
      end;
    end;
  '"': // string
    begin
    tmpPosition := fPosition +1;
    while not ( fText[ tmpPosition ] in [ #0, #13, #10] ) do
      begin
      Inc( tmpPosition );
      if FText[ tmpPosition ] = '"' then
        break;
      end;
    fLength := tmpPosition - fPosition + 1;
    fTokenType := tt_Other;
    end;
  '#' :  // define
    begin
    tmpPosition := fPosition + 1;
    while not ( FText[tmpPosition] in [ #0, #13, #10 ] ) do
      Inc( tmpPosition );
    fTokenType := tt_Define;
    fLength := tmpPosition - fPosition;
    end;
  'a'..'z','A'..'Z','_','-' :
    begin // ident
    tmpPosition := fPosition + 1;
    while FText[tmpPosition] in [ 'a'..'z','A'..'Z','_','-', '0'..'9' ] do
      Inc( tmpPosition );
    fTokenType := tt_Ident;
    fLength := tmpPosition - fPosition;
    if @fOnMacroExpansion <> nil then
      begin
      fColumn := fPosition - FLastEnterPos + 1;
      fToken := OnMacroExpansion( self, Copy( fData, fPosition +1, fLength) );
      exit;
      end;
    end;
  #13 : // new line marker #1
    begin
    tmpPosition := fPosition + 1;
    inc( fRow );
    if FText[ tmpPosition ] = #10 then
      inc( tmpPosition );
    FLastEnterPos := fPosition - 1;
    if @fOnNewLine <> nil then
      OnNewLine( Self, fRow, fPosition - FLastEnterPos + 1,
                 tmpPosition + 1 );
    fTokenType := tt_other;
    fLength := tmpPosition - fPosition;
    end;
  #10 : // new line marker #2
    begin
    tmpPosition := fPosition + 1;
    inc( fRow );
    FLastEnterPos := tmpPosition - 1;
    if @fOnNewLine <> nil then
      OnNewLine( Self, fRow, fPosition - FLastEnterPos + 1,
                 tmpPosition + 1 );
    fTokenType := tt_other;
    fLength := tmpPosition - fPosition;                 
    end;
  else  // other
    begin
    tmpPosition := fPosition + 1;
    while not (FText[tmpPosition] in [#0, '#', '/', '"', 'a'..'z','A'..'Z','_','-']) do
      begin
      if fText[ tmpPosition ] = #13 then
        begin
        inc( fRow );
        if FText[ tmpPosition + 1 ] = #10 then
          inc( tmpPosition );
        FLastEnterPos := tmpPosition + 1;
        if @fOnNewLine <> nil then
          fOnNewLine( Self, fRow, fPosition - FLastEnterPos + 1,
                      tmpPosition + 1 );
        end
      else if FText[tmpPosition] = #10 then
        begin
        inc( fRow );
        FLastEnterPos := tmpPosition + 1;
        if @fOnNewLine <> nil then
          fOnNewLine( Self, fRow, fPosition - FLastEnterPos + 1,
                      tmpPosition + 1 );
        end;
      Inc( tmpPosition );
      end;
    fTokenType := tt_Other;
    fLength := tmpPosition - fPosition;
    end;
end; {case}
fColumn := fPosition - FLastEnterPos + 1;
fToken := Copy( fData, fPosition +1, fLength);
end; {Next}

// -----------------------------------------------------------------------------
//   TPreProcessorDefine
// -----------------------------------------------------------------------------

Constructor TPreProcessorDefine.Create( const aName, aValue : string );
begin
inherited create;
Name := aName;
Value := aValue;
end; {create}

// -----------------------------------------------------------------------------
//   TPreProcessorDefineStates
// -----------------------------------------------------------------------------

constructor TPreProcessorDefineStates.Create;
begin
inherited create;
fItems := Tlist.Create;
end; {create}

destructor TPreProcessorDefineStates.Destroy;
var
  index : integer;
begin
for index := fItems.Count -1 downto 0 do
  TPreProcessorDefineState( FItems[ index ] ).Free;
fItems.Free;
Inherited;
end; {TPreProcessorDefineStates}

function TPreProcessorDefineStates.Add : TPreProcessorDefineState;
begin
assert( self <> nil );
Result := TPreProcessorDefineState.Create;
fItems.Add(Result);
end; {Add}

procedure TPreProcessorDefineStates.Delete( index : Integer );
begin
assert( self <> nil );
TPreProcessorDefineState( fItems[ index ] ).Free;
fItems.Delete( index );
end; {Delete}

function TPreProcessorDefineStates.GetCount: Longint;
begin
assert( self <> nil );
Result := fItems.Count;
end; {GetCount}

function TPreProcessorDefineStates.GetItem( Index : integer ): TPreProcessorDefineState;
begin
assert( self <> nil );
Result := fItems[ Index ];
end; {GetItem}

function TPreProcessorDefineStates.GetWrite: Boolean;
begin
assert( self <> nil );
if FItems.Count <> 0 then
  result := TPreProcessorDefineState( fItems[ fItems.Count -1 ] ).DoWrite
else
  result := true
end; {GetWrite}

// -----------------------------------------------------------------------------
//   TLineInfoList
// -----------------------------------------------------------------------------

constructor TLineInfoList.Create;
begin
inherited Create;
fItems := TList.Create;
end; {Create}

destructor TLineInfoList.Destroy;
begin
Clear;
fItems.Free;
inherited Destroy;
end; {TLineInfoList}

procedure TLineInfoList.Clear;
var
  index : integer;
begin
assert( self <> nil );
for index := FItems.count -1 downto 0 do
  TLineInfo( fItems[ index ] ).Free;
fItems.Clear;
end; {Clear}

function TLineInfoList.Add: TLineInfo;
begin
assert( self <> nil );
Result := TLineInfo.Create;
fItems.Add(Result);
end; {Add}

function TLineInfoList.GetCount: Longint;
begin
assert( self <> nil );
result := fItems.Count;
end; {GetCount}

function TLineInfoList.GetItem( index : Integer) : TLineInfo;
begin
assert( self <> nil );
result := TLineInfo( fItems[ index ] );
end; {GetItem}

function TLineInfoList.GetLineInfo( Position : integer;
                                    var Res: TLineInfoResults): Boolean;
var
  index1 : integer;
  index2 : integer;
  linepos : integer;
  Item: TLineInfo;
begin
assert( self <> nil );
for index1 := FItems.Count -1 downto 0 do
  begin
  Item := FItems[ index1 ];
  if ( Position >= Item.StartPos ) and ( Position < Item.EndPos ) then
    begin
    Res.Name := Item.FileName;
    Position := Position - Item.StartPos;
    Res.Position := Position;
    Res.Column := 1;
    Res.Row := 1;
    LinePos := 0;
    for index2 := 0 to Item.LineOffsetCount -1 do
      begin
      if Position >= Item.LineOffset[ index2 ] then
        linepos := Item.LineOffset[ index2 ]
      else
        begin
        Res.Row := index2; // j -1, but line counting starts at 1
        Res.Column := Position - linepos + 1;
        Break;
        end;
      end;
    Result := true;
    exit;
    end;
  end;
Result := false;
end; {GetLineInfo}

// -----------------------------------------------------------------------------
//   TLineInfo
// -----------------------------------------------------------------------------

constructor TLineInfo.Create;
begin
inherited;
fLineOffsets := TList.Create;
end; {Create}

destructor TLineInfo.Destroy;
begin
FreeAndNil( fLineOffsets );
inherited;
end; {Destroy}

function TLineInfo.GetLineOffset( index : Integer) : Integer;
begin
assert( self <> nil );
Result := Integer( fLineOffsets[ index ] );
end;  {GetLineOffset}

function TLineInfo.GetLineOffsetCount: Longint;
begin
assert( self <> nil );
result := fLineOffsets.Count;
end; {GetLineOffsetCount}

// -----------------------------------------------------------------------------
//   TPreProcessor
// -----------------------------------------------------------------------------

Constructor TPreProcessor.Create;
begin
inherited;
fMaxLevel := 20;
fCurrentLineInfo := TLineInfoList.Create;

fDefineState := TPreProcessorDefineStates.create;

fCurrentDefinesOwner := TObjectList.create(true);
fCurrentDefines := TCaseSensitiveStringList.create;
fCurrentDefines.Sorted := true;
fCurrentDefines.Duplicates := dupError;

fDefinesOwner := TObjectList.create(true);
fDefines := TCaseSensitiveStringList.create;
fDefines.Sorted := true;
fDefines.Duplicates := dupError;

end; {Create}


Destructor TPreProcessor.Destroy;
begin
FreeAndNil( fCurrentLineInfo );
FreeAndNil( fCurrentDefinesOwner );
FreeAndNil( fDefinesOwner );
FreeAndNil( fDefines );
FreeAndNil( fCurrentDefines );
FreeAndNil( fDefineState );
inherited;
end; {destroy}

procedure TPreProcessor.Clear;
begin
fCurrentDefines.Clear;
fCurrentDefinesOwner.Clear;
fCurrentLineInfo.Clear;
fMainFileData := '';
end; {Clear}

Procedure TPreProcessor.ClearDefines;
begin
fDefines.Clear;
fDefinesOwner.Clear;
end; {ClearDefines}

Procedure TPreProcessor.AddPrivateDefine( const Name, value : string );
var
  PreProcessorDefine : TPreProcessorDefine;
  index : integer;
begin
assert( self <> nil );
assert( fCurrentDefines <> nil );
assert( fCurrentDefinesOwner <> nil );
if fCurrentDefines.find( Name, index ) then
  begin
  PreProcessorDefine := fCurrentDefines.objects[  index ] as TPreProcessorDefine;
  PreProcessorDefine.Value := value;
  end
else
  begin
  PreProcessorDefine := TPreProcessorDefine.create( Name,  Trim( value ) );
  fCurrentDefinesOwner.add( PreProcessorDefine );
  fCurrentDefines.AddObject( Name, PreProcessorDefine );
  end;
end; {AddPrivateDefine}

Procedure TPreProcessor.RemovePrivateDefine( const Name : string );
var
  PreProcessorDefine : TPreProcessorDefine;
  index : integer;
begin
assert( self <> nil );
assert( fCurrentDefines <> nil );
if fCurrentDefines.find( Name, index ) then
  begin
  PreProcessorDefine := fCurrentDefines.objects[  index ] as TPreProcessorDefine;
  fCurrentDefines.delete( index );
  if not PreProcessorDefine.fGlobalDefine then
    PreProcessorDefine.Free;
  end;
end; {RemovePrivateDefine}

Function TPreProcessor.PrivateDefineExists( const Name : string ) : boolean;
var
  index : integer;
begin
assert( self <> nil );
assert( fCurrentDefines <> nil );
result := fCurrentDefines.find( Name, index );
end; {PrivateDefineExists}

Procedure TPreProcessor.AddDefine( const Name, value : string );
var
  PreProcessorDefine : TPreProcessorDefine;
  index : integer;
begin
assert( self <> nil );
assert( fDefines <> nil );
assert( fDefinesOwner <> nil );
if fDefines.find( Name, index ) then
  begin
  PreProcessorDefine := fDefines.objects[  index ] as TPreProcessorDefine;
  PreProcessorDefine.Value := value;
  end
else
  begin
  PreProcessorDefine :=  TPreProcessorDefine.create( Name,  Trim( value ) );
  fDefinesOwner.add( PreProcessorDefine );
  fDefines.AddObject( Name, PreProcessorDefine );
  end;
end; {AddDefine}

Procedure TPreProcessor.RemoveDefine( const Name : string );
var
  PreProcessorDefine : TPreProcessorDefine;
  index : integer;
begin
assert( self <> nil );
assert( fDefines <> nil );
if fDefines.find( Name, index ) then
  begin
  // make sure the entry for the private version is removed
  RemovePrivateDefine( Name );
  PreProcessorDefine := fDefines.objects[  index ] as TPreProcessorDefine;
  fDefines.delete( index );
  PreProcessorDefine.Free;
  end;
end; {RemoveDefine}

Function TPreProcessor.DefineExists( const Name : string ) : boolean;
var
  index : integer;
begin
assert( self <> nil );
assert( fDefines <> nil );
result := fDefines.find( Name, index );
end; {DefineExists}

function TPreProcessor.OnMacroExpansion( Sender : TPreProcessorParser;
                                         const MacroName : string ) : string;
var
  index : integer;
begin
assert( self <> nil );
assert( fCurrentDefines <> nil );
if fCurrentDefines.Find( MacroName, index ) then
  begin
  result := TPreProcessorDefine( fCurrentDefines.Objects[  index ] ).Value;
  inc(fAddedPosition, length( result ) - length( MacroName ) );  
  end
else
  result := MacroName;
end; {OnExpandMacro}

procedure TPreProcessor.ParserNewLine( Sender: TPreProcessorParser;
                                       Row, Column, Position: integer);
begin
if CurrentLineInfo.Current >= CurrentLineInfo.Count then exit; //errr ???
with CurrentLineInfo.Items[ CurrentLineInfo.Current ] do
  begin
  Position := Position + FAddedPosition;
  fLineOffsets.Add( Pointer( Position ) );
  end;
end; {ParserNewLine}

procedure TPreProcessor.PreProcess( const Filename: string;
                                    var aOutput: string);
var
  Stream : TMemoryStream;
begin
fAddedPosition := 0;
fCurrentDefines.Assign( fDefines );
Stream := TMemoryStream.Create;
try
  PreProcessFile( 0, '', FileName, Stream );
finally
  Stream.Position := 0;
  SetLength( aOutput, Stream.Size);
  if aOutput <> '' then
    Stream.Read( aOutput[1], Length( aOutput ) );
  Stream.Free;
end;
if fDefineState.Count <> 0 then
  raise EPreProcessorError.Create('Define not closed');
end; {PreProcess}


function UnQuote( const s : string) : string;
var
  len : integer;
begin
result := s;
len := length( result );
if (len > 0) and
   (result[1] = '"') and ( result[len] = '"' )  then
  begin
  setlength( result, len-1);
  delete( result, 1,1 );
  end;
end; {UnQuote}

function FindWhiteSpace( const data : string ) : Integer;
var
  Index : Integer;
begin
for Index := 1 to Length(data) do
  if IsDelimiter( #09' ', Data, Index ) then
    begin
    Result := Index;
    Exit;
    end;
Result := 0;
end; {FindWhiteSpace}

Procedure ParseChunk( data : string;
                      var Command : string;
                      var Params : string );
var
  index : integer;
begin
index := length( data );
while ( index > 0 ) and
      ( data[ index ] in [ #10, #13 ] ) do
  dec( index );
if index <> length( data ) then
  setlength( data, index);
index := FindWhiteSpace( Data );
if index <> 0 then
  begin
  Command := copy( data, 1, index-1 );
  delete( data, 1, index );
  Params := Trim(data);
  end
else
  begin
  Command := Trim(data);
  Params := '';
  end;
end; {ParseChunk}

Procedure ParseParams( Params : string;
                       var Param1 : string;
                       var Param2 : string );
var
  index : integer;
begin
index := length( Params );
while ( index > 0 ) and
      ( Params[ index ] in [ #10, #13 ] ) do
  dec( index );
if index <> length( Params ) then
  setlength( Params, index);
Param1 := '';
Param2 := '';
index := FindWhiteSpace( Params );
if index <> 0 then
  begin
  Param1 := copy( Params, 1, index-1 );
  delete( Params, 1, index );
  Param2 := Trim(Params);
  end
else
  begin
  Param1 := Trim(Params);
  Param2 := '';
  end;
end; {ParseParams}

procedure TPreProcessor.PreProcessFile( Level: Integer;
                                        const OrgFileName: string;
                                        FileName: string;
                                        Dest: TStream);
var
  Parser : TPreProcessorParser;
  data : string;
  item : TLineInfo;
  chunk : string;
  Command : string;
  Params, Param1, Param2 : string;
  current : integer;
  index : integer;
  aDefineState : TPreProcessorDefineState;
  NoPreprocessorError : boolean;
  Res : TLineInfoResults;
begin
if Level > MaxLevel then
  raise EPreProcessorError.Create( 'Too many nested include files while processing '+
                                   FileName+' from '+OrgFileName );
Parser := TPreProcessorParser.Create;
try
  Parser.OnNewLine := ParserNewLine;
  Parser.OnMacroExpansion := OnMacroExpansion;
  if FileName = MainFileName then
    data := MainFileData
  else if (@OnNeedFile = nil) or
          (not OnNeedFile( Self, OrgFileName, FileName, data ) ) then
    raise EPreProcessorError.Create('Unable to find file '''+filename+'''');
  Item := CurrentLineInfo.Add;
  current := CurrentLineInfo.Count -1;
  CurrentLineInfo.Current := current;
  Item.FStartPos := Dest.Position;
  Item.FFileName := FileName;
  Parser.SetText( data );
  try
  while Parser.TokenType <> tt_EOF do
    begin
    chunk := Parser.Token;
    if Parser.TokenType = tt_Define then
      begin
      // delete the '#'
      Delete( chunk, 1, 1);
      ParseChunk( chunk, Command, Params );
      NoPreprocessorError := true;
      if Command = 'include' then
        begin
        if fDefineState.DoWrite then
          begin
          fAddedPosition := 0;
          PreProcessFile( Level + 1, FileName, UnQuote( Params ), Dest);
          CurrentLineInfo.Current := current;
          fAddedPosition := Dest.Position - Item.StartPos - Parser.Position;
          end;
        end
      else if Command = 'define' then
        begin
        ParseParams( Params, Param1, Param2 );
        AddPrivateDefine( Param1, Param2 )
        end
      else if Command = 'undef' then
        RemovePrivateDefine( Param1 )
      else if Command = 'ifndef' then
        fDefineState.Add.DoWrite := FCurrentDefines.IndexOf( Uppercase( Params ) ) = -1
      else if Command = 'ifdef' then
        fDefineState.Add.DoWrite := FCurrentDefines.IndexOf( Uppercase( Params ) ) <> -1
      else if Command = 'endif' then
        begin
        if FDefineState.Count = 0 then   
          raise EPreProcessorError.Create( 'No #ifdef for #endif' );
        fDefineState.Delete( fDefineState.Count -1 ); // remove define from list
        end
      else if Command = 'else' then
        begin
        if FDefineState.Count = 0 then
          raise Exception.Create( 'No #ifdef for #else' );
        aDefineState := fDefineState[ fDefineState.Count ];
        if aDefineState.InElse then
          raise EPreProcessorError.Create( 'Can''t use #else twice' );
        aDefineState.FInElse := True;
        aDefineState.DoWrite := not aDefineState.DoWrite;
        end
      else
        NoPreprocessorError := (@fOnCommand = nil) or OnCommand( self, command, Params );
      if not NoPreprocessorError then
        raise EPreProcessorError.CreateFmt( 'Unknown compiler directives : %s',
                                            [ command ]);

      end;
    if (not FDefineState.DoWrite) or (Parser.TokenType = tt_Define) then
      begin
      SetLength( chunk, Length( Parser.Token ) );
      for index := length( chunk ) downto 1 do
        chunk[ index ] := #32; // space
      end;
    if chunk <> '' then
      Dest.Write( chunk[1], length(chunk));
    Parser.Next;
    end;
  Item.FEndPos := Dest.Position;   
  except
    on E : Exception do
      begin
      Item.FEndPos := Dest.Position+1; 
      if CurrentLineInfo.GetLineInfo( Parser.Position, Res ) then
        chunk := Format( '"%s"(%d:%d) : %s',
                         [ FileName, Res.Row, Res.Column, E.message ] )
      else
        chunk := Format( '"%s"(%d:%d) : %s',
                         [ FileName, Parser.Row, Parser.Column, E.message ] );
      raise EPreProcessorError.Create( chunk ) at ExceptAddr;
      end;
  end;
finally
  Parser.free;
end;
end; {PreProcessFile}



end.
